package librarymanagementsystem;

import java.io.*;
import java.util.*;

/**
 * Represents the library of a Library Management System.
 * 
 * @author Pawprint Studios
 */
public class Library 
{

    public String libraryName;
    public String libraryAddress;
    public static double dailyLateFine = 0.25;
    public int maxLoans = 30;

    public ArrayList<Book> libraryBooks = new ArrayList<>();
    protected ArrayList<GeneralUser> libraryUsers = new ArrayList<>();

    protected long highestUserID;
    protected long highestBookID;

    private String bookLocation;
    private String userLocation;

    /**
     * Represents the databases of books and users.
     */
    public Library() 
    {
        this("../resources/Books.csv", "../resources/Users.csv");
    }

    /**
     * Represents where the databases are held.
     * 
     * @param bookLocation represents the location of where the book database
     *                     is being held.
     * @param userLocation represents the location of where the book database
     *                     is being held.
     */
    public Library(String bookLocation, String userLocation) 
    {
        this.bookLocation = bookLocation;
        this.userLocation = userLocation;

        updateBooksArray();
        updateUsersArray();

        highestUserID = libraryUsers.size() + 1;
        highestBookID = libraryBooks.size() + 1;
    }

    
    
    // *** Accessors. ***

    /**
     * Get the library name.
     * 
     * @return libraryname
     */
    public String getLibraryName() 
    {
        return libraryName;
    }

    /**
     * Get the library address.
     * 
     * @return libraryAddress
     */
    public String getLibraryAddress() 
    {
        return libraryAddress;
    }

    /**
     * Get the daily late fine.
     * 
     * @return dailyLateFine
     */
    public double getDailyLateFine() 
    {
        return dailyLateFine;
    }

    /**
     * Get the Highest User ID.
     * 
     * @return highestUserID
     */
    public long getHighestUserID() 
    {
        return highestUserID;
    }

    public long getHighestBookID() 
    {
        return highestBookID;
    }

    /**
     * Get all the library books within the library.
     * 
     * @return printList
     */
    public String getLibraryBooks() 
    {
        String printList = "";
        for (Book printBook : libraryBooks) {
            printList += printBook;
        }
        return printList;
    }

    /**
     * Get all the users within the Library.
     * 
     * @return printList
     */
    public String getLibraryUsers() 
    {
        String printList = "";
        for (GeneralUser printUser : libraryUsers) {
            printList += printUser;
        }
        return printList;
    }

    
    
    // *** Mutators. ***

    /**
     * Set name of the Library management system.
     * 
     * @param name represents the name of the library.
     */
    protected void setLibraryName(String name) 
    {
        libraryName = name;
    }

    /**
     * Set the address for the Library.
     * 
     * @param address represents the address of the library.
     */
    protected void setLibraryAddress(String address) 
    {
        libraryAddress = address;
    }

    /**
     * Set how much a user will get fined for each day a book isn't returned.
     * 
     * @param dailyFine represents the daily late fine for a user.
     */
    protected void setDailyLateFine(Double dailyFine) 
    {
        dailyLateFine = dailyFine;
    }

    /**
     * Update the database for the book and adds books.
     */
    public void updateBooksArray() 
    {
        libraryBooks.clear();

        try (Scanner bookScan = new Scanner(new File(bookLocation))) 
        {
            bookScan.nextLine();
            while (bookScan.hasNextLine()) 
            {
                String[] bookParts = bookScan.nextLine().trim().split(",");

                String title = bookParts[0];
                String author = bookParts[1];
                int year = Integer.parseInt(bookParts[2]);
                String publisher = bookParts[3];
                String publisherDate = bookParts[4];
                int edition = Integer.parseInt(bookParts[5]);
                long isbn = Long.parseLong(bookParts[6]);
                int numCopies = Integer.parseInt(bookParts[7]);
                int numCheckedCopies = Integer.parseInt(bookParts[8]);

                long idNum = Long.parseLong(bookParts[9]);
                double price = Double.parseDouble(bookParts[10]);

                ArrayList<Long> currentHolders = new ArrayList<>(numCopies);
                for (int i = 11; i < 10 + numCheckedCopies; i++) 
                {
                    currentHolders.add(Long.parseLong(bookParts[i]));
                }

                libraryBooks.add(new Book(title, author, year, publisher,
                        publisherDate, edition, isbn, numCopies, numCheckedCopies,
                        idNum, price, currentHolders));
            }
        } 
        catch (FileNotFoundException e) 
        {
            System.out.println("Could not find file: " + bookLocation);
            System.exit(1);
        }
    }

    /**
     * Update the user database and adds users.
     */
    public void updateUsersArray() 
    {
        libraryUsers.clear();

        try (Scanner fileScan = new Scanner(new File(userLocation))) 
        {
            fileScan.nextLine();
            while (fileScan.hasNextLine()) 
            {
                String[] userParts = fileScan.nextLine().trim().split(",");

                String username = userParts[0];
                String password = userParts[1];
                long userId = Long.parseLong(userParts[2]);
                String userType = userParts[3];

                String name = userParts[4];
                String email = userParts[5];
                long phoneNumber = Long.parseLong(userParts[6]);
                
                if (userType.equals("Patron")) 
                {

                    int numLoans = Integer.parseInt(userParts[7]);

                    ArrayList<Long> currentLoans = new ArrayList(maxLoans);
                    for (int i = 8; i <= 7 + numLoans; i++) {
                        currentLoans.add(Long.parseLong(userParts[i]));
                    }

                    libraryUsers.add(new GeneralUser(username, password, userId,
                            userType, name, email, phoneNumber, numLoans,
                            currentLoans ));
                } 
                else 
                {
                    libraryUsers.add(new GeneralUser(username, password, userId,
                            userType, name, email, phoneNumber));
                
                }
            }
        } 
        catch (FileNotFoundException e) 
        {
            System.out.println("Could not find file: " + userLocation);
            System.exit(2);
            // Ryan Note: I still don't know what to put in the catch blocks.
        }
    }

    public void updateBooksFile(String location)
    {
        try (PrintWriter outFile = new PrintWriter(
            new BufferedWriter(new FileWriter(location)))) 
        {
            outFile.print("Title,Author,Year,Publisher,PublishDate,ISBN," 
                + "NumCopies,NumChecked,IdNum,Price,CurrentHolders");
            for (Book libBook : libraryBooks)
            {
                outFile.print(libBook.toString());
            }
        } 
        catch (IOException e) 
        {
            System.out.println(e);
        }
    }
    
    public void updateUsersFile(String location)
    {
        try (PrintWriter outFile = new PrintWriter(
            new BufferedWriter(new FileWriter(location)))) 
        {
            outFile.print("Username,Password,UserID,UserType,Name,Email," 
                + "PhoneNumber,NumLoans,CurrentLoans");
            for (GeneralUser libUser : libraryUsers)
            {
                outFile.print(libUser.toString());
            }
        } 
        catch (IOException e) 
        {
            System.out.println(e);
        }
    }
    
    
    // *** Abilities that all users can access. ***

    /**
     * Verify the login of a user.
     * 
     * @param username represents the username of a user.
     * @param password represents the password of a user.
     * @return loginUser
     */
    public GeneralUser verifyLogin(String username, String password) 
    {
        GeneralUser loginUser = null;

        for (GeneralUser arrayUser : libraryUsers) 
        {
            if (username.equalsIgnoreCase(arrayUser.getUsername())
                    && password.equals(arrayUser.getPassword())) 
            {
                loginUser = arrayUser;
            }
        }

        return loginUser;
    }

    /**
     * Allow guest to apply for Membership of a Patron
     * 
     * @param username represents the username given by user.
     * @param password represents the password given by user.
     * @return successfulAplication represents if user was created
     */
    public boolean applyMembership(String username, String password, 
        String name, String email, long phoneNumber) 
    {
        boolean successfulAplication = false;
        if (findUser(username) == null) 
        {
            libraryUsers.add(new GeneralUser(username, password, highestUserID, 
                "Patron", name, email, phoneNumber, 0, 
                new ArrayList<>(maxLoans)));
            updateUsersFile(userLocation);
            highestUserID++;
            successfulAplication = true;
        } 
        else 
        {
            System.out.println("A user with that username already exists!");
        }
        return successfulAplication;
    }

    /**
     * Search the library for a book/media.
     * 
     * @param bookSearch represents the name of the book being searched.
     */
    public ArrayList<Book> searchLibrary(String bookSearch) 
    {
        ArrayList<Book> foundBooks = new ArrayList<>();

        String searchBook = bookSearch.toLowerCase();
        for (Book arrayBook : libraryBooks) 
        {
            if (arrayBook.getTitle().toLowerCase().contains(searchBook)
                || arrayBook.getAuthor().toLowerCase().contains(searchBook)) 
            {
                foundBooks.add(arrayBook);
            }
        }

        return foundBooks;
    }

    
    
    // *** Abilities that Patrons can access. ***

    public Book findBook(long libIdNum) 
    {
        for (Book arrayBook : libraryBooks) 
        {
            if (arrayBook.getIdNum() == libIdNum) 
            {
                return arrayBook;
            }
        }
        return null;
    }
    
    public Book findBook(String bookTitle) 
    {
        for (Book arrayBook : libraryBooks) 
        {
            if (arrayBook.getTitle().contains(bookTitle)) 
            {
                return arrayBook;
            }
        }
        return null;
    }

    public String borrowBook(long userIdNum, Book bookSearch) 
    {
        String successfulBorrow = "";
        GeneralUser user = findUser(userIdNum);
        long bookIdNum = bookSearch.getIdNum();
        
        if (!user.currentLoans.contains(bookIdNum))
        {
            if (user.getNumLoans() < maxLoans)
            {
                user.loanBook(bookIdNum);

                bookSearch.loanBook(userIdNum);

                updateBooksFile(bookLocation);
                updateUsersFile(userLocation);
                
                successfulBorrow = "Borrowed";
            } 
            else 
            {
                successfulBorrow = "You already have " + maxLoans + " books.";
            }
        }
        else
        {
            successfulBorrow = "You already have " + bookSearch.getTitle()
                + " checked out.";
        }
        return successfulBorrow; 
    }

    public ArrayList<Book> viewLoans(long userIdNum) 
    {
        ArrayList<Book> loanedBooks = new ArrayList<>();
        for (long lo : findUser(userIdNum).getCurrentLoans()) 
        {
            loanedBooks.add(findBook(lo));
        }
        return loanedBooks;
    }

    public String returnBook(long userIdNum, long bookIdNum) 
    {
        String successfulReturn = "";
        GeneralUser user = findUser(userIdNum);
        if (user.currentLoans.contains(bookIdNum))
        {
            user.returnBook(bookIdNum);
            findBook(bookIdNum).returnBook(userIdNum);
            updateBooksFile(bookLocation);
            updateUsersFile(userLocation);
            successfulReturn = "Returned";
        }
        else
        {
            successfulReturn = "You don't have a copy of " + findBook(bookIdNum)
                .getTitle() + " to return.";
        }
        return successfulReturn;
    }


    
    
    // *** Abilities that Librarians can access. ***

    /**
     * Add book into array list.
     * 
     * @param addingBook represents the book being added to the array list.
     */
    public void addBook(Book addingBook) 
    {
        libraryBooks.add(addingBook);
        updateBooksFile(bookLocation);
        highestBookID++;
    }

    /**
     * Retire/remove book from array list.
     * 
     * @param retireIdNum represents the ID number of the book being retired.
     */
    public void retireBook(long retireIdNum) 
    {
        libraryBooks.remove(findBook(retireIdNum));
        updateBooksFile(bookLocation);
    }

    /**
     * Find user from user array list.
     * 
     * @param userSearch represents what the user searched.
     * @return
     */
    public GeneralUser findUser(String userSearch) 
    {

        for (GeneralUser arrayUser : libraryUsers) 
        {
            if (userSearch.equalsIgnoreCase(arrayUser.getUsername())) 
            {
                return arrayUser;
            }
        }

        return null;
    }

    /**
     * Find user from user array list.
     * 
     * @param userIdSearch represents what the user searched.
     * @return GeneralUser
     */
    public GeneralUser findUser(long userIdSearch) 
    {

        for (GeneralUser arrayUser : libraryUsers) 
        {
            if (userIdSearch == arrayUser.getUserId()) 
            {
                return arrayUser;
            }
        }

        return null;
    }

    // *** Abilities that Admins can access. ***

    /**
    * Remove users.
    * 
    * @param User 
    */
    public void removeUser(long userId)
    {
        libraryUsers.remove(findUser(userId));
        updateUsersFile(userLocation);
    }
    
    /**
     * Backup database.
     * @param backupLocation
     */
    public void backupDatabase()
    {
        updateBooksFile("../backups/Books.csv");
        updateUsersFile("../backups/Users.csv");
    }

}
